@@ -1,4 +1,4 @@
-from typing import Optional
+from typing import Optional, Union, Text
 
 import torch
 from einops import rearrange
@@ -18,8 +18,13 @@ class SpeakerSegmentation:
         self.formatter = TemporalFeatureFormatter()
 
     @staticmethod
-    def from_pyannote(model, device: Optional[torch.device] = None) -> 'SpeakerSegmentation':
-        return SpeakerSegmentation(SegmentationModel.from_pyannote(model), device)
+    def from_pyannote(
+        model,
+        use_hf_token: Union[Text, bool, None] = True,
+        device: Optional[torch.device] = None
+    ) -> 'SpeakerSegmentation':
+        seg_model = SegmentationModel.from_pyannote(model, use_hf_token)
+        return SpeakerSegmentation(seg_model, device)
 
     def __call__(self, waveform: TemporalFeatures) -> TemporalFeatures:
         """
