@@ -9,8 +9,8 @@ from optuna.trial import Trial, FrozenTrial
 from tqdm import trange, tqdm
 
 from .audio import FilePath
-from .benchmark import Benchmark
-from .blocks import PipelineConfig, OnlineSpeakerDiarization
+from .blocks import BasePipelineConfig, PipelineConfig, OnlineSpeakerDiarization
+from .inference import Benchmark
 
 
 @dataclass
@@ -39,12 +39,17 @@ class Optimizer:
     def __init__(
         self,
         speech_path: Union[Text, Path],
-        reference_path: Optional[Union[Text, Path]],
+        reference_path: Union[Text, Path],
         study_or_path: Union[FilePath, Study],
         batch_size: int = 32,
+        pipeline_class: type = OnlineSpeakerDiarization,
         hparams: Optional[Sequence[HyperParameter]] = None,
-        base_config: Optional[PipelineConfig] = None,
+        base_config: Optional[BasePipelineConfig] = None,
+        do_kickstart_hparams: bool = True,
     ):
+        self.pipeline_class = pipeline_class
+        # FIXME can we run this benchmark in parallel?
+        #  Currently it breaks the trial progress bar
         self.benchmark = Benchmark(
             speech_path,
             reference_path,
@@ -52,11 +57,24 @@ class Optimizer:
             show_report=False,
             batch_size=batch_size,
         )
-        self.base_config = PipelineConfig() if base_config is None else base_config
+
+        self.base_config = base_config
+        self.do_kickstart_hparams = do_kickstart_hparams
+        if self.base_config is None:
+            self.base_config = PipelineConfig()
+            self.do_kickstart_hparams = False
+
         self.hparams = hparams
         if self.hparams is None:
             self.hparams = [TauActive, RhoUpdate, DeltaNew]
 
+        # Make sure hyper-parameters exist in the configuration class given
+        possible_hparams = vars(self.base_config)
+        for param in self.hparams:
+            msg = f"Hyper-parameter {param.name} not found " \
+                  f"in configuration {self.base_config.__class__.__name__}"
+            assert param.name in possible_hparams, msg
+
         self._progress: Optional[tqdm] = None
 
         if isinstance(study_or_path, Study):
@@ -104,11 +122,11 @@ class Optimizer:
         if trial.should_prune():
             raise TrialPruned()
 
-        # Instantiate pipeline with the new configuration
-        pipeline = OnlineSpeakerDiarization(PipelineConfig(**trial_config))
+        # Instantiate the new configuration for the trial
+        config = self.base_config.__class__(**trial_config)
 
         # Run pipeline over the dataset
-        report = self.benchmark(pipeline)
+        report = self.benchmark(self.pipeline_class, config)
 
         # Extract DER from report
         return report.loc["TOTAL", "diarization error rate"]["%"]
@@ -121,4 +139,10 @@ class Optimizer:
             if self.study.trials:
                 last_trial = self.study.trials[-1].number
             self._progress.set_description(f"Trial {last_trial + 1}")
+        # Start with base config hyper-parameters if config was given
+        if self.do_kickstart_hparams:
+            self.study.enqueue_trial({
+                param.name: getattr(self.base_config, param.name)
+                for param in self.hparams
+            }, skip_if_exists=True)
         self.study.optimize(self.objective, num_iter, callbacks=[self._callback])
